@@ -1,7 +1,7 @@
 /*
  * Copyright © 2009 CNRS
- * Copyright © 2009-2014 Inria.  All rights reserved.
- * Copyright © 2009-2010 Université Bordeaux 1
+ * Copyright © 2009-2017 Inria.  All rights reserved.
+ * Copyright © 2009-2010 Université Bordeaux
  * Copyright © 2009-2011 Cisco Systems, Inc.  All rights reserved.
  * See COPYING in top-level directory.
  */
@@ -60,7 +60,7 @@ hwloc_synthetic_process_level_indexes(struct hwloc_synthetic_backend_data_s *dat
   unsigned long length = curlevel->index_string_length;
   unsigned *array = NULL;
   struct hwloc_synthetic_intlv_loop_s * loops = NULL;
-  unsigned long i;
+  size_t i;
 
   if (!attr)
     return;
@@ -81,7 +81,7 @@ hwloc_synthetic_process_level_indexes(struct hwloc_synthetic_backend_data_s *dat
       unsigned idx = strtoul(attr, (char **) &next, 10);
       if (next == attr) {
 	if (verbose)
-	  fprintf(stderr, "Failed to read synthetic index #%lu at '%s'\n", i, attr);
+	  fprintf(stderr, "Failed to read synthetic index #%lu at '%s'\n", (unsigned long) i, attr);
 	goto out_with_array;
       }
 
@@ -89,7 +89,7 @@ hwloc_synthetic_process_level_indexes(struct hwloc_synthetic_backend_data_s *dat
       if (i != total-1) {
 	if (*next != ',') {
 	  if (verbose)
-	    fprintf(stderr, "Missing comma after synthetic index #%lu at '%s'\n", i, attr);
+	    fprintf(stderr, "Missing comma after synthetic index #%lu at '%s'\n", (unsigned long) i, attr);
 	  goto out_with_array;
 	}
 	attr = next+1;
@@ -137,6 +137,11 @@ hwloc_synthetic_process_level_indexes(struct hwloc_synthetic_backend_data_s *dat
 	    fprintf(stderr, "Failed to read synthetic index interleaving loop '%s' without number before '*'\n", tmp);
 	  goto out_with_loops;
 	}
+	if (!step) {
+	  if (verbose)
+	    fprintf(stderr, "Invalid interleaving loop with step 0 at '%s'\n", tmp);
+	  goto out_with_loops;
+	}
 	tmp2++;
 	nb = (unsigned) strtol(tmp2, &tmp3, 0);
 	if (tmp3 == tmp2 || (*tmp3 && *tmp3 != ':' && *tmp3 != ')' && *tmp3 != ' ')) {
@@ -144,6 +149,11 @@ hwloc_synthetic_process_level_indexes(struct hwloc_synthetic_backend_data_s *dat
 	    fprintf(stderr, "Failed to read synthetic index interleaving loop '%s' without number between '*' and ':'\n", tmp);
 	  goto out_with_loops;
 	}
+	if (!nb) {
+	  if (verbose)
+	    fprintf(stderr, "Invalid interleaving loop with number 0 at '%s'\n", tmp2);
+	  goto out_with_loops;
+	}
 	loops[cur_loop].step = step;
 	loops[cur_loop].nb = nb;
 	if (step < minstep)
@@ -172,6 +182,11 @@ hwloc_synthetic_process_level_indexes(struct hwloc_synthetic_backend_data_s *dat
 	    fprintf(stderr, "Failed to read synthetic index interleaving loop type '%s'\n", tmp);
 	  goto out_with_loops;
 	}
+	if (type == HWLOC_OBJ_MISC || type == HWLOC_OBJ_BRIDGE || type == HWLOC_OBJ_PCI_DEVICE || type == HWLOC_OBJ_OS_DEVICE) {
+	  if (verbose)
+	    fprintf(stderr, "Misc object type disallowed in synthetic index interleaving loop type '%s'\n", tmp);
+	  goto out_with_loops;
+	}
 	for(i=0; i<curleveldepth; i++) {
 	  if (type != data->level[i].type)
 	    continue;
@@ -183,7 +198,7 @@ hwloc_synthetic_process_level_indexes(struct hwloc_synthetic_backend_data_s *dat
 	      && cachetypeattr != (hwloc_obj_cache_type_t) -1
 	      && cachetypeattr != data->level[i].cachetype)
 	    continue;
-	  loops[cur_loop].level_depth = i;
+	  loops[cur_loop].level_depth = (unsigned)i;
 	  break;
 	}
 	if (i == curleveldepth) {
@@ -219,11 +234,14 @@ hwloc_synthetic_process_level_indexes(struct hwloc_synthetic_backend_data_s *dat
 
 	loops[cur_loop].step = step;
 	loops[cur_loop].nb = nb;
+	assert(nb);
+	assert(step);
 	if (step < minstep)
 	  minstep = step;
 	nbs *= nb;
       }
     }
+    assert(nbs);
 
     if (nbs != total) {
       /* one loop of total/nbs steps is missing, add it if it's just the smallest one */
@@ -276,6 +294,29 @@ hwloc_synthetic_process_level_indexes(struct hwloc_synthetic_backend_data_s *dat
   return;
 }
 
+static hwloc_uint64_t
+hwloc_synthetic_parse_memory_attr(const char *attr, const char **endp)
+{
+  const char *endptr;
+  hwloc_uint64_t size;
+  size = strtoull(attr, (char **) &endptr, 0);
+  if (!hwloc_strncasecmp(endptr, "TB", 2)) {
+    size <<= 40;
+    endptr += 2;
+  } else if (!hwloc_strncasecmp(endptr, "GB", 2)) {
+    size <<= 30;
+    endptr += 2;
+  } else if (!hwloc_strncasecmp(endptr, "MB", 2)) {
+    size <<= 20;
+    endptr += 2;
+  } else if (!hwloc_strncasecmp(endptr, "kB", 2)) {
+    size <<= 10;
+    endptr += 2;
+  }
+  *endp = endptr;
+  return size;
+}
+
 static int
 hwloc_synthetic_parse_level_attrs(const char *attrs, const char **next_posp,
 				  struct hwloc_synthetic_level_data_s *curlevel,
@@ -285,7 +326,7 @@ hwloc_synthetic_parse_level_attrs(const char *attrs, const char **next_posp,
   const char *next_pos;
   hwloc_uint64_t memorysize = 0;
   const char *index_string = NULL;
-  unsigned long index_string_length = 0;
+  size_t index_string_length = 0;
 
   next_pos = (const char *) strchr(attrs, ')');
   if (!next_pos) {
@@ -297,10 +338,10 @@ hwloc_synthetic_parse_level_attrs(const char *attrs, const char **next_posp,
 
   while (')' != *attrs) {
     if (HWLOC_OBJ_CACHE == type && !strncmp("size=", attrs, 5)) {
-      memorysize = strtoull(attrs+5, (char**) &attrs, 0);
+      memorysize = hwloc_synthetic_parse_memory_attr(attrs+5, &attrs);
 
     } else if (HWLOC_OBJ_CACHE != type && !strncmp("memory=", attrs, 7)) {
-      memorysize = strtoull(attrs+7, (char**) &attrs, 0);
+      memorysize = hwloc_synthetic_parse_memory_attr(attrs+7, &attrs);
 
     } else if (!strncmp("indexes=", attrs, 8)) {
       index_string = attrs+8;
@@ -327,7 +368,7 @@ hwloc_synthetic_parse_level_attrs(const char *attrs, const char **next_posp,
 
   curlevel->memorysize = memorysize;
   curlevel->index_string = index_string;
-  curlevel->index_string_length = index_string_length;
+  curlevel->index_string_length = (unsigned long)index_string_length;
   *next_posp = next_pos+1;
   return 0;
 }
@@ -346,7 +387,7 @@ hwloc_backend_synthetic_init(struct hwloc_synthetic_backend_data_s *data,
   int nb_machine_levels = 0, nb_node_levels = 0;
   int nb_pu_levels = 0;
   int verbose = 0;
-  char *env = getenv("HWLOC_SYNTHETIC_VERBOSE");
+  const char *env = getenv("HWLOC_SYNTHETIC_VERBOSE");
   int err;
   unsigned long totalarity = 1;
 
@@ -387,6 +428,12 @@ hwloc_backend_synthetic_init(struct hwloc_synthetic_backend_data_s *data,
 	errno = EINVAL;
 	goto error;
       }
+      if (type == HWLOC_OBJ_SYSTEM || type == HWLOC_OBJ_MISC || type == HWLOC_OBJ_BRIDGE || type == HWLOC_OBJ_PCI_DEVICE || type == HWLOC_OBJ_OS_DEVICE) {
+	if (verbose)
+	  fprintf(stderr, "Synthetic string with disallowed object type at '%s'\n", pos);
+	errno = EINVAL;
+	goto error;
+      }
 
       next_pos = strchr(pos, ':');
       if (!next_pos) {
@@ -408,6 +455,12 @@ hwloc_backend_synthetic_init(struct hwloc_synthetic_backend_data_s *data,
       errno = EINVAL;
       goto error;
     }
+    if (!item) {
+      if (verbose)
+	fprintf(stderr,"Synthetic string with disallow 0 number of objects at '%s'\n", pos);
+      errno = EINVAL;
+      goto error;
+    }
     data->level[count-1].arity = (unsigned)item;
 
     totalarity *= item;
@@ -450,6 +503,19 @@ hwloc_backend_synthetic_init(struct hwloc_synthetic_backend_data_s *data,
 
     type = curlevel->type;
 
+    if (i == count-1 && type != HWLOC_OBJ_TYPE_UNKNOWN && type != HWLOC_OBJ_PU) {
+      if (verbose)
+	fprintf(stderr, "Synthetic string cannot use non-PU type for last level\n");
+      errno = EINVAL;
+      return -1;
+    }
+    if (i != count-1 && type == HWLOC_OBJ_PU) {
+      if (verbose)
+	fprintf(stderr, "Synthetic string cannot use PU type for non-last level\n");
+      errno = EINVAL;
+      return -1;
+    }
+
     if (type == HWLOC_OBJ_TYPE_UNKNOWN) {
       if (i == count-1)
 	type = HWLOC_OBJ_PU;
@@ -457,12 +523,11 @@ hwloc_backend_synthetic_init(struct hwloc_synthetic_backend_data_s *data,
 	switch (data->level[i+1].type) {
 	case HWLOC_OBJ_PU: type = HWLOC_OBJ_CORE; break;
 	case HWLOC_OBJ_CORE: type = HWLOC_OBJ_CACHE; break;
-	case HWLOC_OBJ_CACHE: type = HWLOC_OBJ_SOCKET; break;
-	case HWLOC_OBJ_SOCKET: type = HWLOC_OBJ_NODE; break;
-	case HWLOC_OBJ_NODE:
-	case HWLOC_OBJ_GROUP: type = HWLOC_OBJ_GROUP; break;
+	case HWLOC_OBJ_CACHE: type = HWLOC_OBJ_PACKAGE; break;
+	case HWLOC_OBJ_PACKAGE: type = HWLOC_OBJ_NUMANODE; break;
+	case HWLOC_OBJ_NUMANODE:
 	case HWLOC_OBJ_MACHINE:
-	case HWLOC_OBJ_MISC: type = HWLOC_OBJ_MISC; break;
+	case HWLOC_OBJ_GROUP: type = HWLOC_OBJ_GROUP; break;
 	default:
 	  assert(0);
 	}
@@ -479,7 +544,7 @@ hwloc_backend_synthetic_init(struct hwloc_synthetic_backend_data_s *data,
       case HWLOC_OBJ_GROUP:
 	group_depth++;
 	break;
-      case HWLOC_OBJ_NODE:
+      case HWLOC_OBJ_NUMANODE:
 	nb_node_levels++;
 	break;
       case HWLOC_OBJ_MACHINE:
@@ -548,7 +613,7 @@ hwloc_backend_synthetic_init(struct hwloc_synthetic_backend_data_s *data,
 	  curlevel->memorysize = 256*1024 << (2*curlevel->depth);
       }
 
-    } else if (type == HWLOC_OBJ_NODE && !curlevel->memorysize) {
+    } else if (type == HWLOC_OBJ_NUMANODE && !curlevel->memorysize) {
       /* 1GB in memory nodes. */
       curlevel->memorysize = 1024*1024*1024;
     }
@@ -575,23 +640,16 @@ hwloc_synthetic__post_look_hooks(struct hwloc_synthetic_level_data_s *curlevel,
 				 hwloc_obj_t obj)
 {
   switch (obj->type) {
-  case HWLOC_OBJ_MISC:
-    break;
   case HWLOC_OBJ_GROUP:
     obj->attr->group.depth = curlevel->depth;
     break;
-  case HWLOC_OBJ_BRIDGE:
-  case HWLOC_OBJ_PCI_DEVICE:
-  case HWLOC_OBJ_OS_DEVICE:
-    abort();
-    break;
   case HWLOC_OBJ_SYSTEM:
     break;
   case HWLOC_OBJ_MACHINE:
     break;
-  case HWLOC_OBJ_NODE:
+  case HWLOC_OBJ_NUMANODE:
     break;
-  case HWLOC_OBJ_SOCKET:
+  case HWLOC_OBJ_PACKAGE:
     break;
   case HWLOC_OBJ_CACHE:
     obj->attr->cache.depth = curlevel->depth;
@@ -603,6 +661,10 @@ hwloc_synthetic__post_look_hooks(struct hwloc_synthetic_level_data_s *curlevel,
     break;
   case HWLOC_OBJ_PU:
     break;
+  case HWLOC_OBJ_BRIDGE:
+  case HWLOC_OBJ_PCI_DEVICE:
+  case HWLOC_OBJ_OS_DEVICE:
+  case HWLOC_OBJ_MISC:
   case HWLOC_OBJ_TYPE_MAX:
     /* Should never happen */
     assert(0);
@@ -640,22 +702,13 @@ hwloc__look_synthetic(struct hwloc_topology *topology,
 
   /* pre-hooks */
   switch (type) {
-    case HWLOC_OBJ_MISC:
-      break;
     case HWLOC_OBJ_GROUP:
       break;
-    case HWLOC_OBJ_SYSTEM:
-    case HWLOC_OBJ_BRIDGE:
-    case HWLOC_OBJ_PCI_DEVICE:
-    case HWLOC_OBJ_OS_DEVICE:
-      /* Shouldn't happen.  */
-      abort();
-      break;
     case HWLOC_OBJ_MACHINE:
       break;
-    case HWLOC_OBJ_NODE:
+    case HWLOC_OBJ_NUMANODE:
       break;
-    case HWLOC_OBJ_SOCKET:
+    case HWLOC_OBJ_PACKAGE:
       break;
     case HWLOC_OBJ_CACHE:
       break;
@@ -663,6 +716,11 @@ hwloc__look_synthetic(struct hwloc_topology *topology,
       break;
     case HWLOC_OBJ_PU:
       break;
+    case HWLOC_OBJ_SYSTEM:
+    case HWLOC_OBJ_BRIDGE:
+    case HWLOC_OBJ_PCI_DEVICE:
+    case HWLOC_OBJ_OS_DEVICE:
+    case HWLOC_OBJ_MISC:
     case HWLOC_OBJ_TYPE_MAX:
       /* Should never happen */
       assert(0);
@@ -682,7 +740,7 @@ hwloc__look_synthetic(struct hwloc_topology *topology,
       hwloc__look_synthetic(topology, data, level + 1, obj->cpuset);
   }
 
-  if (type == HWLOC_OBJ_NODE) {
+  if (type == HWLOC_OBJ_NUMANODE) {
     obj->nodeset = hwloc_bitmap_alloc();
     hwloc_bitmap_set(obj->nodeset, os_index);
   }
@@ -812,7 +870,7 @@ static int hwloc_topology_export_synthetic_indexes(struct hwloc_topology * topol
   unsigned total = topology->level_nbobjects[depth];
   unsigned step = 1;
   unsigned nr_loops = 0;
-  struct hwloc_synthetic_intlv_loop_s *loops = NULL;
+  struct hwloc_synthetic_intlv_loop_s *loops = NULL, *tmploops;
   hwloc_obj_t cur;
   unsigned i, j;
   ssize_t tmplen = buflen;
@@ -839,9 +897,10 @@ static int hwloc_topology_export_synthetic_indexes(struct hwloc_topology * topol
 	break;
 
     nr_loops++;
-    loops = realloc(loops, nr_loops*sizeof(*loops));
-    if (!loops)
+    tmploops = realloc(loops, nr_loops*sizeof(*loops));
+    if (!tmploops)
       goto exportall;
+    loops = tmploops;
     loops[nr_loops-1].step = i;
     loops[nr_loops-1].nb = j;
     step *= j;
@@ -869,7 +928,7 @@ static int hwloc_topology_export_synthetic_indexes(struct hwloc_topology * topol
     }
     ret += res;
     if (res >= tmplen)
-      res = tmplen>0 ? tmplen - 1 : 0;
+      res = tmplen>0 ? (int)tmplen - 1 : 0;
     tmp += res;
     tmplen -= res;
   }
@@ -892,7 +951,7 @@ static int hwloc_topology_export_synthetic_indexes(struct hwloc_topology * topol
       return -1;
     ret += res;
     if (res >= tmplen)
-      res = tmplen>0 ? tmplen - 1 : 0;
+      res = tmplen>0 ? (int)tmplen - 1 : 0;
     tmp += res;
     tmplen -= res;
     cur = cur->next_cousin;
@@ -920,7 +979,7 @@ static int hwloc_topology_export_synthetic_obj_attr(struct hwloc_topology * topo
 	     prefix, (unsigned long long) obj->memory.local_memory);
     prefix = separator;
   }
-  if (obj->type == HWLOC_OBJ_PU || obj->type == HWLOC_OBJ_NODE) {
+  if (obj->type == HWLOC_OBJ_PU || obj->type == HWLOC_OBJ_NUMANODE) {
     hwloc_obj_t cur = obj;
     while (cur) {
       if (cur->os_index != cur->logical_index) {
@@ -940,7 +999,7 @@ static int hwloc_topology_export_synthetic_obj_attr(struct hwloc_topology * topo
       return -1;
     ret += res;
     if (res >= tmplen)
-      res = tmplen>0 ? tmplen - 1 : 0;
+      res = tmplen>0 ? (int)tmplen - 1 : 0;
     tmp += res;
     tmplen -= res;
 
@@ -950,7 +1009,7 @@ static int hwloc_topology_export_synthetic_obj_attr(struct hwloc_topology * topo
 	return -1;
       ret += res;
       if (res >= tmplen)
-	res = tmplen>0 ? tmplen - 1 : 0;
+	res = tmplen>0 ? (int)tmplen - 1 : 0;
       tmp += res;
       tmplen -= res;
 
@@ -959,7 +1018,7 @@ static int hwloc_topology_export_synthetic_obj_attr(struct hwloc_topology * topo
 	return -1;
       ret += res;
       if (res >= tmplen)
-	res = tmplen>0 ? tmplen - 1 : 0;
+	res = tmplen>0 ? (int)tmplen - 1 : 0;
       tmp += res;
       tmplen -= res;
     }
@@ -978,7 +1037,7 @@ hwloc_topology_export_synthetic(struct hwloc_topology * topology,
   ssize_t tmplen = buflen;
   char *tmp = buffer;
   int res, ret = 0;
-   int arity;
+  unsigned arity;
   const char * separator = " ";
   const char * prefix = "";
 
@@ -1014,7 +1073,7 @@ hwloc_topology_export_synthetic(struct hwloc_topology * topology,
     if (ret > 0)
       prefix = separator;
     if (res >= tmplen)
-      res = tmplen>0 ? tmplen - 1 : 0;
+      res = tmplen>0 ? (int)tmplen - 1 : 0;
     tmp += res;
     tmplen -= res;
   }
@@ -1034,7 +1093,7 @@ hwloc_topology_export_synthetic(struct hwloc_topology * topology,
       return -1;
     ret += res;
     if (res >= tmplen)
-      res = tmplen>0 ? tmplen - 1 : 0;
+      res = tmplen>0 ? (int)tmplen - 1 : 0;
     tmp += res;
     tmplen -= res;
 
@@ -1045,7 +1104,7 @@ hwloc_topology_export_synthetic(struct hwloc_topology * topology,
 	return -1;
       ret += res;
       if (res >= tmplen)
-	res = tmplen>0 ? tmplen - 1 : 0;
+	res = tmplen>0 ? (int)tmplen - 1 : 0;
       tmp += res;
       tmplen -= res;
     }
